import ujson as json
from asgiref.sync import async_to_sync
from mock import Mock, patch
from tornado.testing import AsyncHTTPTestCase

from consoleme.lib.self_service.models import (
    SelfServiceTypeaheadModel,
    SelfServiceTypeaheadModelArray,
)
from consoleme.models import AwsResourcePrincipalModel
from tests.conftest import MockBaseHandler, create_future


class TestTypeAheadHandler(AsyncHTTPTestCase):
    def get_app(self):
        from consoleme.routes import make_app

        return make_app(jwt_validator=lambda x: {})

    def setUp(self):
        super(TestTypeAheadHandler, self).setUp()

    @patch(
        "consoleme.handlers.v2.typeahead.ResourceTypeAheadHandlerV2.authorization_flow",
        MockBaseHandler.authorization_flow,
    )
    def test_typeahead_get(self):
        from consoleme.config import config

        headers = {
            config.get("auth.user_header_name"): "user@github.com",
            config.get("auth.groups_header_name"): "groupa,groupb,groupc",
        }
        from consoleme.lib.redis import RedisHandler

        red = RedisHandler().redis_sync()
        red.hmset(
            "AWSCONFIG_RESOURCE_CACHE",
            {
                "arn:aws:ec2:us-west-2:123456789013:security-group/12345": "{}",
                "arn:aws:sqs:us-east-1:123456789012:rolequeue": "{}",
                "arn:aws:sns:us-east-1:123456789012:roletopic": "{}",
                "arn:aws:iam::123456789012:role/role": "{}",
            },
        )
        # Return all the things
        response = self.fetch(
            "/api/v2/typeahead/resources", method="GET", headers=headers
        )
        self.assertEqual(response.code, 200)
        responseJSON = json.loads(response.body)

        self.assertEqual(len(responseJSON), 4)
        # Filter for a specific query
        response = self.fetch(
            "/api/v2/typeahead/resources?typeahead=role", method="GET", headers=headers
        )
        self.assertEqual(response.code, 200)
        responseJSON = json.loads(response.body)
        self.assertEqual(len(responseJSON), 3)

        # Filter for a specific limit
        response = self.fetch(
            "/api/v2/typeahead/resources?typeahead=role&limit=1",
            method="GET",
            headers=headers,
        )
        self.assertEqual(response.code, 200)
        responseJSON = json.loads(response.body)
        self.assertEqual(len(responseJSON), 1)

        # Filter for a specific account
        response = self.fetch(
            "/api/v2/typeahead/resources?account_id=123456789013",
            method="GET",
            headers=headers,
        )
        self.assertEqual(response.code, 200)
        responseJSON = json.loads(response.body)
        self.assertEqual(len(responseJSON), 1)

        # Filter for a specific resource type
        response = self.fetch(
            "/api/v2/typeahead/resources?resource_type=sqs",
            method="GET",
            headers=headers,
        )
        self.assertEqual(response.code, 200)
        responseJSON = json.loads(response.body)
        self.assertEqual(len(responseJSON), 1)

        # filter for region
        response = self.fetch(
            "/api/v2/typeahead/resources?region=us-east-1",
            method="GET",
            headers=headers,
        )
        self.assertEqual(response.code, 200)
        responseJSON = json.loads(response.body)
        self.assertEqual(len(responseJSON), 2)

        # multifilter
        response = self.fetch(
            "/api/v2/typeahead/resources?region=us-east-1&account_id=123456789012&typeahead=role&limit=5",
            method="GET",
            headers=headers,
        )
        self.assertEqual(response.code, 200)
        responseJSON = json.loads(response.body)
        self.assertEqual(len(responseJSON), 2)

    def test_cache_self_service_template_and_typeahead(self):
        from consoleme.lib.templated_resources import (
            TemplatedFileModelArray,
            TemplateFile,
        )

        mock_template_file_model_array = TemplatedFileModelArray(
            templated_resources=[
                TemplateFile(
                    name="fake_test_template_1",
                    repository_name="fake_repo",
                    owner="fake_owner",
                    include_accounts=["fake_account_1"],
                    exclude_accounts=None,
                    number_of_accounts=1,
                    resource="path/to/file.yaml",
                    file_path="path/to/file.yaml",
                    web_path="http://github.example.com/fake_repo/browse/master/path/to/file.yaml",
                    resource_type="iam_role",
                    template_language="honeybee",
                )
            ]
        )

        mock_template_typeahead_model = SelfServiceTypeaheadModel(
            details_endpoint="/api/v2/templated_resource/fake_repo/path/to/file.yaml",
            display_text="fake_test_template_1",
            icon="users",
            number_of_affected_resources=1,
            principal={
                "principal_type": "HoneybeeAwsResourceTemplate",
                "repository_name": "fake_repo",
                "resource_identifier": "path/to/file.yaml",
                "resource_url": "http://github.example.com/fake_repo/browse/master/path/to/file.yaml",
            },
        )

        patch_cache_resource_templates_for_repository = patch(
            "consoleme.lib.templated_resources.cache_resource_templates_for_repository",
            Mock(return_value=create_future(mock_template_file_model_array)),
        )

        # Cache resource templates, but let's not go down the rabbit hole of trying to mock a Git repo
        patch_cache_resource_templates_for_repository.start()
        from consoleme.lib.templated_resources import cache_resource_templates

        result = async_to_sync(cache_resource_templates)()
        patch_cache_resource_templates_for_repository.stop()
        self.assertEqual(result, mock_template_file_model_array)

        # Retrieve cached resource templates and ensure it is correct
        from consoleme.lib.templated_resources import retrieve_cached_resource_templates

        result = async_to_sync(retrieve_cached_resource_templates)()
        self.assertEqual(result, mock_template_file_model_array)

        # Cache and verify Self Service Typeahead
        from consoleme.lib.self_service.typeahead import cache_self_service_typeahead

        result = async_to_sync(cache_self_service_typeahead)()
        self.assertIsInstance(result, SelfServiceTypeaheadModelArray)
        self.assertGreater(len(result.typeahead_entries), 15)
        expected_entry = SelfServiceTypeaheadModel(
            account="default_account",
            details_endpoint="/api/v2/roles/123456789012/RoleNumber5",
            display_text="RoleNumber5",
            icon="user",
            number_of_affected_resources=1,
            principal=AwsResourcePrincipalModel(
                principal_type="AwsResource",
                principal_arn="arn:aws:iam::123456789012:role/RoleNumber5",
            ),
        )
        # Pre-existing role is in results
        self.assertIn(expected_entry, result.typeahead_entries)
        # HB template is in results
        self.assertIn(mock_template_typeahead_model, result.typeahead_entries)

        # Now let's mock the web requests
        from consoleme.config import config

        headers = {
            config.get("auth.user_header_name"): "user@github.com",
            config.get("auth.groups_header_name"): "groupa,groupb,groupc",
        }

        response = self.fetch(
            "/api/v2/templated_resource/fake_repo/path/to/file.yaml",
            method="GET",
            headers=headers,
        )
        self.assertEqual(response.code, 200)
        response_body = json.loads(response.body)
        self.assertEqual(
            response_body,
            {
                "name": "fake_test_template_1",
                "owner": "fake_owner",
                "include_accounts": ["fake_account_1"],
                "exclude_accounts": None,
                "number_of_accounts": 1,
                "resource": "path/to/file.yaml",
                "resource_type": "iam_role",
                "repository_name": "fake_repo",
                "template_language": "honeybee",
                "web_path": "http://github.example.com/fake_repo/browse/master/path/to/file.yaml",
                "file_path": "path/to/file.yaml",
                "content": None,
            },
        )



import sys
from typing import List

import boto3
from asgiref.sync import sync_to_async

from consoleme.config import config
from consoleme.lib.generic import generate_html, get_principal_friendly_name
from consoleme.lib.groups import get_group_url
from consoleme.lib.plugins import get_plugin_by_name
from consoleme.models import ExtendedRequestModel, RequestStatus

stats = get_plugin_by_name(config.get("plugins.metrics", "default_metrics"))()
log = config.get_logger()


async def send_email(
    to_addresses: List[str],
    subject: str,
    body: str,
    sending_app: str = "consoleme",
    region: str = config.get("ses.region", "us-west-2"),
    charset: str = "UTF-8",
) -> None:
    client = boto3.client(
        "ses", region_name=region, **config.get("boto3.client_kwargs", {})
    )
    sender = config.get(f"ses.{sending_app}.sender")
    log_data = {
        "to_user": to_addresses,
        "region": region,
        "function": f"{__name__}.{sys._getframe().f_code.co_name}",
        "sender": sender,
        "subject": subject,
    }

    if not config.get("ses.arn"):
        log.error(
            {
                **log_data,
                "error": "Configuration value for `ses.arn` is not defined. Unable to send e-mail.",
            }
        )
        return

    if not sender:
        log.error(
            {
                **log_data,
                "error": f"Configuration value for `ses.{sending_app}.sender` is not defined. Unable to send e-mail.",
            }
        )
        return

    if config.get("development") and config.get("ses.override_receivers_for_dev"):
        log_data["original_to_addresses"] = to_addresses
        log_data["message"] = "Overriding to_address"
        to_addresses = config.get("ses.override_receivers_for_dev")
        log_data["new_to_addresses"] = to_addresses
        log.debug(log_data)

    # Handle non-list recipients
    if not isinstance(to_addresses, list):
        to_addresses = [to_addresses]

    try:
        response = await sync_to_async(client.send_email)(
            Destination={"ToAddresses": to_addresses},  # This should be a list
            Message={
                "Body": {
                    "Html": {"Charset": charset, "Data": body},
                    "Text": {"Charset": charset, "Data": body},
                },
                "Subject": {"Charset": charset, "Data": subject},
            },
            Source=sender,
            SourceArn=config.get("ses.arn"),
        )
    # Display an error if something goes wrong.
    except Exception:
        stats.count("lib.ses.error")
        log_data["message"] = "Exception sending email"
        log.error(log_data, exc_info=True)
    else:
        stats.count("lib.ses.success")
        log_data["message"] = "Email sent successfully"
        log_data["response"] = response["MessageId"]
        log.debug(log_data)


async def send_access_email_to_user(
    user: str,
    group: str,
    updated_by: str,
    status: str,
    request_url: str,
    group_url: str,
    reviewer_comments: None = None,
    sending_app: str = "consoleme",
) -> None:
    app_name = config.get(f"ses.{sending_app}.name", sending_app)
    subject = f"{app_name}: Request for group {group} has been {status}"
    to_addresses = [user, updated_by]
    group_link = f"<a href={group_url}>{group}</a>"
    message = f"Your request for group {group_link} has been {status} by {updated_by}."
    if status == "approved":
        message += " Please allow up to 30 minutes for your group to propagate. "

    reviewer_comments_section = ""
    if reviewer_comments:
        reviewer_comments_section = f"Reviewer Comments: {reviewer_comments}"
    body = f"""<html>
    <head>
    <meta http-equiv="content-type" content="text/html; charset=UTF-8">
    <title>Request Status</title>
    </head>
    <body>
    {message} <br>
    {reviewer_comments_section} <br>
    See your request here: {request_url}.<br>
    <br>
    {config.get('ses.support_reference', '')}
    <meta http-equiv="content-type" content="text/html; charset=UTF-8">
    </body>
    </html>"""
    await send_email(to_addresses, subject, body, sending_app=sending_app)


async def send_request_created_to_user(
    user, group, updated_by, status, request_url, sending_app="consoleme"
):
    app_name = config.get(f"ses.{sending_app}.name", sending_app)
    subject = f"{app_name}: Request for group {group} has been created"
    to_addresses = [user, updated_by]
    message = f"Your request for group {group} has been created."
    if status == "approved":
        message += " Please allow up to 30 minutes for your group to propagate. "
    body = f"""<html>
    <head>
    <meta http-equiv="content-type" content="text/html; charset=UTF-8">
    <title>Request Status</title>
    </head>
    <body>
    {message} <br>
    <br>
    See your request here: {request_url}.<br>
    <br>
    {config.get('ses.support_reference', '')}
    <meta http-equiv="content-type" content="text/html; charset=UTF-8">
    </body>
    </html>"""
    await send_email(to_addresses, subject, body, sending_app=sending_app)


async def send_request_to_secondary_approvers(
    secondary_approvers,
    group,
    request_url,
    pending_requests_url,
    sending_app="consoleme",
):
    app_name = config.get(f"ses.{sending_app}.name", sending_app)
    subject = f"{app_name}: A request for group {group} requires your approval"
    to_addresses = secondary_approvers
    message = f"A request for group {group} requires your approval."
    body = f"""<html>
        <head>
        <meta http-equiv="content-type" content="text/html; charset=UTF-8">
        <title>Request Status</title>
        </head>
        <body>
        {message} <br>
        <br>
        See the request here: {request_url}.<br>
        <br>
        You can find all pending requests waiting your approval here: {pending_requests_url}. <br>
        <br>
        {config.get('ses.support_reference', '')}
        <meta http-equiv="content-type" content="text/html; charset=UTF-8">
        </body>
        </html>"""
    await send_email(to_addresses, subject, body, sending_app=sending_app)


async def send_group_modification_notification(
    groups, to_address, sending_app="consoleme"
):
    """
    Send an email containing group changes to a notification address

    Example of `groups` dict:
    {
        "awesome_group_1@netflix.com": [
            {"name": "tswift@netflix.com", "type": "USER"},
            {"name": "agrande@netflix.com", "type": "USER"},
        ],
        "awesome_group_2@netflix.com": [
            {"name": "lizzo@netflix.com", "type": "USER"},
            {"name": "beilish@netflix.com", "type": "USER"},
        ],
    }

    :param groups: map of groups and added members
    :type groups: dict
    :param to_address: recipient of notification email
    :type to_address: str
    :param sending_app: name of application
    :type sending_app: str
    """
    app_name = config.get(f"ses.{sending_app}.name", sending_app)
    subject = f"{app_name}: Groups modified"
    message = f"""Groups modified in {app_name}.<br>
    You or a group you belong to are configured to receive a notification when new members are added to this group.<br>
    Admins may click the group link below to view and modify this configuration."""
    added_members_snippet = ""
    for group, added_members in groups.items():
        group_url = get_group_url(group)
        group_link = f"<a href={group_url}>{group}</a>"
        if added_members:
            added_members_snippet += f"""<b>Users added to {group_link}</a></b>: <br>
            {generate_html(added_members)}<br>
            """
    body = f"""<html>
        <head>
        <meta http-equiv="content-type" content="text/html; charset=UTF-8">
        </head>
        <body>
         {message}<br>
         <br>
         {added_members_snippet}<br>
        <br>
        <br>
        {config.get('ses.support_reference', '')}
        <meta http-equiv="content-type" content="text/html; charset=UTF-8">
        </body>
        </html>"""
    await send_email(to_address, subject, body, sending_app=sending_app)


async def send_new_aws_groups_notification(
    to_addresses, new_aws_groups, sending_app="consoleme"
):
    app_name = config.get(f"ses.{sending_app}.name", sending_app)
    subject = f"{app_name}: New AWS groups detected"
    message = """New AWS login groups were created.<br>
    ConsoleMe is configured to send notifications when new AWS-related google groups are detected.
    This is to detect any accidentally or maliciously created google groups.<br>"""
    added_groups_snippet = ""
    if new_aws_groups:
        added_groups_snippet = f"""<b>New groups</b>: <br>
        {generate_html({"New Groups": new_aws_groups})}<br>
        """
    body = f"""<html>
        <head>
        <meta http-equiv="content-type" content="text/html; charset=UTF-8">
        </head>
        <body>
         {message}<br>
         <br>
         {added_groups_snippet}<br>
        <br>
        <br>
        {config.get('ses.support_reference', '')}
        <meta http-equiv="content-type" content="text/html; charset=UTF-8">
        </body>
        </html>"""
    await send_email(to_addresses, subject, body, sending_app=sending_app)


async def send_policy_request_status_update(
    request, policy_change_uri, sending_app="consoleme"
):
    app_name = config.get(f"ses.{sending_app}.name", sending_app)
    subject = f"{app_name}: Policy change request for {request['arn']} has been {request['status']}"
    if request["status"] == "pending":
        subject = (
            f"{app_name}: Policy change request for {request['arn']} has been created"
        )
    to_addresses = [request.get("username")]
    message = (
        f"A policy change request for {request['arn']} has been {request['status']}"
    )
    if request["status"] == "pending":
        message = f"A policy change request for {request['arn']} has been created."
    if {request["status"]} == "approved":
        message += " and committed"
        subject += " and committed"
    body = f"""<html>
            <head>
            <meta http-equiv="content-type" content="text/html; charset=UTF-8">
            <title>Policy Change Request Status Change</title>
            </head>
            <body>
            {message} <br>
            <br>
            See the request here: {policy_change_uri}.<br>
            <br>
            <br>
            {config.get('ses.support_reference', '')}
            <meta http-equiv="content-type" content="text/html; charset=UTF-8">
            </body>
            </html>"""
    await send_email(to_addresses, subject, body, sending_app=sending_app)


async def send_policy_request_status_update_v2(
    extended_request: ExtendedRequestModel, policy_change_uri, sending_app="consoleme"
):
    app_name = config.get(f"ses.{sending_app}.name", sending_app)
    to_addresses = [extended_request.requester_email]
    principal = await get_principal_friendly_name(extended_request.principal)

    if extended_request.request_status == RequestStatus.pending:
        subject = f"{app_name}: Policy change request for {principal} has been created"
        message = f"A policy change request for {principal} has been created."
        # This is a new request, also send email to application admins
        to_addresses.append(config.get("application_admin"))
    else:
        subject = (
            f"{app_name}: Policy change request for {principal} has been "
            f"updated to {extended_request.request_status.value}"
        )
        message = (
            f"A policy change request for {principal} "
            f"has been updated to {extended_request.request_status.value}"
        )

    if extended_request.request_status == RequestStatus.approved:
        message += " and committed"
        subject += " and committed"
    body = f"""<html>
            <head>
            <meta http-equiv="content-type" content="text/html; charset=UTF-8">
            <title>Policy Change Request Status Change</title>
            </head>
            <body>
            {message} <br>
            <br>
            See the request here: {policy_change_uri}.<br>
            <br>
            <br>
            {config.get('ses.support_reference', '')}
            <meta http-equiv="content-type" content="text/html; charset=UTF-8">
            </body>
            </html>"""
    await send_email(to_addresses, subject, body, sending_app=sending_app)


async def send_new_comment_notification(
    extended_request: ExtendedRequestModel,
    to_addresses,
    user,
    policy_change_uri,
    sending_app="consoleme",
):
    app_name = config.get(f"ses.{sending_app}.name", sending_app)
    principal = await get_principal_friendly_name(extended_request.principal)
    subject = f"{app_name}: A new comment has been added to Policy Change request for {principal}"
    message = f"A new comment has been added to the policy change request for {principal} by {user}"
    body = f"""<html>
                <head>
                <meta http-equiv="content-type" content="text/html; charset=UTF-8">
                <title>Policy Change Request Comment Notification</title>
                </head>
                <body>
                {message} <br>
                <br>
                See the request here: {policy_change_uri}.<br>
                <br>
                <br>
                {config.get('ses.support_reference', '')}
                <meta http-equiv="content-type" content="text/html; charset=UTF-8">
                </body>
                </html>"""
    await send_email(to_addresses, subject, body, sending_app=sending_app)



import sys
from typing import List, Optional

import boto3
import sentry_sdk
import ujson as json
from botocore.exceptions import ClientError
from cloudaux.aws.sts import boto3_cached_conn

from consoleme.config import config
from consoleme.exceptions.exceptions import MissingConfigurationValue
from consoleme.lib.plugins import get_plugin_by_name

log = config.get_logger()
stats = get_plugin_by_name(config.get("plugins.metrics", "default_metrics"))()


def query(
    query: str, use_aggregator: bool = True, account_id: Optional[str] = None
) -> List:
    resources = []
    if use_aggregator:
        config_client = boto3.client(
            "config", region_name=config.region, **config.get("boto3.client_kwargs", {})
        )
        configuration_aggregator_name: str = config.get(
            "aws_config.configuration_aggregator.name"
        ).format(region=config.region)
        if not configuration_aggregator_name:
            raise MissingConfigurationValue("Invalid configuration for aws_config")
        response = config_client.select_aggregate_resource_config(
            Expression=query,
            ConfigurationAggregatorName=configuration_aggregator_name,
            Limit=100,
        )
        for r in response.get("Results", []):
            resources.append(json.loads(r))
        while response.get("NextToken"):
            response = config_client.select_aggregate_resource_config(
                Expression=query,
                ConfigurationAggregatorName=configuration_aggregator_name,
                Limit=100,
                NextToken=response["NextToken"],
            )
            for r in response.get("Results", []):
                resources.append(json.loads(r))
        return resources
    else:  # Don't use Config aggregator and instead query all the regions on an account
        session = boto3.Session()
        available_regions = config.get("aws_config.available_regions", [])
        if not available_regions:
            available_regions = session.get_available_regions("config")
        excluded_regions = config.get(
            "api_protect.exclude_regions",
            ["af-south-1", "ap-east-1", "ap-northeast-3", "eu-south-1", "me-south-1"],
        )
        regions = [x for x in available_regions if x not in excluded_regions]
        for region in regions:
            config_client = boto3_cached_conn(
                "config",
                account_number=account_id,
                assume_role=config.get("policies.role_name"),
                region=region,
                sts_client_kwargs=dict(
                    region_name=config.region,
                    endpoint_url=config.get(
                        "aws.sts_endpoint_url", "https://sts.{region}.amazonaws.com"
                    ).format(region=config.region),
                ),
                client_kwargs=config.get("boto3.client_kwargs", {}),
            )
            try:
                response = config_client.select_resource_config(
                    Expression=query, Limit=100
                )
                for r in response.get("Results", []):
                    resources.append(json.loads(r))
                # Query Config for a specific account in all regions we care about
                while response.get("NextToken"):
                    response = config_client.select_resource_config(
                        Expression=query, Limit=100, NextToken=response["NextToken"]
                    )
                    for r in response.get("Results", []):
                        resources.append(json.loads(r))
            except ClientError as e:
                log.error(
                    {
                        "function": f"{__name__}.{sys._getframe().f_code.co_name}",
                        "message": "Failed to query AWS Config",
                        "query": query,
                        "use_aggregator": use_aggregator,
                        "account_id": account_id,
                        "region": region,
                        "error": str(e),
                    },
                    exc_info=True,
                )
                sentry_sdk.capture_exception()
        return resources



import ujson as json

from consoleme.config import config
from consoleme.lib.account_indexers.aws_organizations import (
    retrieve_accounts_from_aws_organizations,
)
from consoleme.lib.account_indexers.current_account import retrieve_current_account
from consoleme.lib.account_indexers.local_config import retrieve_accounts_from_config
from consoleme.lib.account_indexers.swag import retrieve_accounts_from_swag
from consoleme.lib.cache import (
    retrieve_json_data_from_redis_or_s3,
    store_json_results_in_redis_and_s3,
)
from consoleme.lib.plugins import get_plugin_by_name
from consoleme.models import CloudAccountModelArray

log = config.get_logger(__name__)
auth = get_plugin_by_name(config.get("plugins.auth", "default_auth"))()
stats = get_plugin_by_name(config.get("plugins.metrics", "default_metrics"))()


async def cache_cloud_accounts() -> CloudAccountModelArray:
    """
    Gets Cloud Account Information from either ConsoleMe's configuration, AWS Organizations, or Swag,
    depending on configuration
    :return:
    """
    account_mapping = None
    # Get the accounts
    if config.get("cache_cloud_accounts.from_aws_organizations"):
        account_mapping = await retrieve_accounts_from_aws_organizations()
    elif config.get("cache_cloud_accounts.from_swag"):
        account_mapping = await retrieve_accounts_from_swag()
    elif config.get("cache_cloud_accounts.from_config", True):
        account_mapping = await retrieve_accounts_from_config()

    if not account_mapping or not account_mapping.accounts:
        account_mapping = await retrieve_current_account()

    account_id_to_name = {}

    for account in account_mapping.accounts:
        account_id_to_name[account.id] = account.name

    redis_key = config.get(
        "cache_cloud_accounts.redis.key.all_accounts_key", "ALL_AWS_ACCOUNTS"
    )

    s3_bucket = None
    s3_key = None
    if config.region == config.get("celery.active_region", config.region) or config.get(
        "environment"
    ) in ["dev", "test"]:
        s3_bucket = config.get("cache_cloud_accounts.s3.bucket")
        s3_key = config.get(
            "cache_cloud_accounts.s3.file", "cache_cloud_accounts/accounts_v1.json.gz"
        )
    # Store full mapping of the model
    # We want to pass a dict to store_json_results_in_redis_and_s3, but the problem is account_mapping.dict()
    # includes pydantic objects that cannot be dumped to json without passing a special JSON encoder for the
    # Pydantic type, hence the usage of json.loads(account_mapping.json())
    await store_json_results_in_redis_and_s3(
        json.loads(account_mapping.json()),
        redis_key=redis_key,
        s3_bucket=s3_bucket,
        s3_key=s3_key,
    )

    return account_mapping


async def get_cloud_account_model_array(
    status="active", environment=None, force_sync=False
):
    redis_key = config.get(
        "cache_cloud_accounts.redis.key.all_accounts_key", "ALL_AWS_ACCOUNTS"
    )
    accounts = await retrieve_json_data_from_redis_or_s3(redis_key, default={})
    if force_sync or not accounts or not accounts.get("accounts"):
        # Force a re-sync and then retry
        await cache_cloud_accounts()
        accounts = await retrieve_json_data_from_redis_or_s3(redis_key, default={})
    all_accounts = CloudAccountModelArray.parse_obj(accounts)
    filtered_accounts = CloudAccountModelArray(accounts=[])
    for account in all_accounts.accounts:
        if status and account.status.value != status:
            continue
        if environment and not account.environment:
            # if we are looking to filter on environment, and account doesn't contain environment information
            continue
        if environment and account.environment.value != environment:
            continue
        filtered_accounts.accounts.append(account)
    return filtered_accounts


async def get_account_id_to_name_mapping(
    status="active", environment=None, force_sync=False
):
    redis_key = config.get(
        "cache_cloud_accounts.redis.key.all_accounts_key", "ALL_AWS_ACCOUNTS"
    )
    accounts = await retrieve_json_data_from_redis_or_s3(redis_key, default={})
    if force_sync or not accounts or not accounts.get("accounts"):
        # Force a re-sync and then retry
        await cache_cloud_accounts()
        accounts = await retrieve_json_data_from_redis_or_s3(
            redis_key,
            s3_bucket=config.get("cache_cloud_accounts.s3.bucket"),
            s3_key=config.get(
                "cache_cloud_accounts.s3.file",
                "cache_cloud_accounts/accounts_v1.json.gz",
            ),
            default={},
        )

    account_id_to_name = {}
    for account in accounts.get("accounts", []):
        if status and account.get("status") != status:
            continue
        if environment and account.get("environment") != environment:
            continue
        account_id_to_name[account["id"]] = account["name"]
    return account_id_to_name



import sys
import threading
from random import random
from typing import Dict, Optional

import aiozipkin as az
from pydantic import BaseModel

from consoleme.config import config

SERVER = "SERVER"
log = config.get_logger()


class ConsoleMeTracerObject(BaseModel):
    headers: Dict
    tracer: az.tracer.Tracer
    primary_span: az.tracer.Span

    class Config:
        arbitrary_types_allowed = True


class ConsoleMeTracer:
    def __init__(self):
        self.spans = {}
        self.headers = {}
        self.log_data = {}
        self.tracer = None
        self.primary_span = None

    def trace_calls(self, frame, event, arg):
        """
        Uses sys.settrace / threading.settrace hooks to create / finish Zipkin spans for function calls,
        exceptions, and returns. In a production environment, sys.settrace should be sampled.

        :param frame: Python Frame Object - https://docs.python.org/3/reference/datamodel.html#types
        :param event: A string describing the type of event (call, return, exception, etc) -
            https://docs.python.org/3/library/sys.html#sys.settrace
        :param arg: The return value or exception raised by the function, if relevant
        :return:
        """

        # Skip anything that's not a call, return, or exception
        if event not in ["call", "return", "exception"]:
            return

        # Set a reproducible pan name
        span_name = f"{frame.f_code.co_filename}-{frame.f_code.co_name}-{frame.f_back.f_code.co_filename}"

        # Skip tracing functions outside of core ConsoleMe by default
        in_scope_function_calls = config.get(
            "tracing.in_scope_function_calls", ["/consoleme/"]
        )
        if not (
            any(x in frame.f_code.co_filename for x in in_scope_function_calls)
            or any(
                x in frame.f_back.f_code.co_filename for x in in_scope_function_calls
            )
        ):
            if self.spans.get(span_name):
                self.spans[span_name].finish()
            return

        if self.spans.get(span_name):
            if event == "exception":
                if isinstance(arg[1], Exception) and arg[2]:  # Ensure traceback exists
                    self.spans.get(span_name).tag(
                        "error", arg[1]
                    )  # Record exception string as a tag
            self.spans[span_name].finish()
            return
        span = self.tracer.new_child(self.primary_span.context)  # Start a child span

        if event == "exception":
            if isinstance(arg[1], Exception) and arg[2]:  # Ensure traceback exists
                span.tag("error", arg[1])  # Record exception string as a tag

        span.kind(SERVER)
        span.start()
        span.name(frame.f_code.co_name)

        # Tag the span with context
        span.tag("FUNCTION", frame.f_code.co_name)
        span.tag("FUNCTION_LINE_NUM", frame.f_lineno)
        span.tag("FILENAME", frame.f_code.co_filename)
        span.tag("CALLER_FILENAME", frame.f_back.f_code.co_filename)
        span.tag("CALLER_LINE_NUM", frame.f_back.f_lineno)

        self.spans[span_name] = span
        frame.f_trace_lines = False
        frame.f_trace = self.trace_calls

    async def configure_tracing(
        self, span_name, span_kind=SERVER, tags=None, annotations=None
    ) -> Optional[ConsoleMeTracerObject]:
        if not config.get("tracing.enabled", False):
            return

        if not random() * 100 <= config.get("tracing.sample_rate", 0.1):  # nosec
            return

        if not tags:
            tags = []
        if not annotations:
            annotations = []
        zipkin_address = config.get(
            "tracing.zipkin_address", "http://127.0.0.1:9411/api/v2/spans"
        ).format(region=config.region, environment=config.get("environment"))
        endpoint = az.create_endpoint(
            config.get("tracing.application_name", "consoleme")
        )
        # The tracer's sample rate is 100% because we are pre-sampling our requests
        self.tracer = await az.create(zipkin_address, endpoint, sample_rate=1.0)
        self.primary_span = self.tracer.new_trace(sampled=True)
        self.headers = self.primary_span.context.make_headers()
        self.log_data = {
            "function": f"{__name__}.{self.__class__.__name__}.{sys._getframe().f_code.co_name}",
            "message": "Starting trace",
            "trace_id": self.primary_span.context.trace_id,
            "zipkin_address": zipkin_address,
            "tags": tags,
            "hostname": config.hostname,
        }
        log.debug(self.log_data)

        self.primary_span.kind(span_kind)
        self.primary_span.start()
        self.primary_span.name(span_name)
        self.primary_span.tag("HOSTNAME", config.hostname)

        for k, v in tags.items():
            self.primary_span.tag(k, v)
        for annotation in annotations:
            self.primary_span.annotate(annotation)

        # Configure sys/threading.settrace to use our trace_calls function for tracing
        # Note: This is expensive, and should definitely not run for every request
        sys.settrace(self.trace_calls)
        threading.settrace(self.trace_calls)
        return ConsoleMeTracerObject(
            primary_span=self.primary_span, tracer=self.tracer, headers=self.headers
        )

    async def disable_tracing(self):
        if not config.get("tracing.enabled", False):
            return
        self.log_data["message"] = "disabling tracing"
        log.debug(self.log_data)
        sys.settrace(None)
        threading.settrace(None)

    async def set_additional_tags(self, tags):
        if self.primary_span:
            for k, v in tags.items():
                self.primary_span.tag(k, v)

    async def finish_spans(self):
        """
        Closes all of the spans and the tracer. Our trace should be visible on ZipKin after this function
        completes.
        :return:
        """
        # Finish any nested spans that are still open.
        # We iterate through the dictionary of spans in a safe manner
        # to avoid a "Dictionary changed size during iteration" error as other asynchronous
        # callers modify self.spans
        if not config.get("tracing.enabled", False):
            return
        self.log_data["message"] = "finishing spans"
        log.debug(self.log_data)
        for span_id in list(self.spans):
            span = self.spans.get(span_id)
            if span:
                span.finish()

        # Finish primary span
        if self.primary_span:
            self.primary_span.finish()

        # Close the tracer
        if self.tracer:
            await self.tracer.close()



import os
from shutil import rmtree

import distutils.cmd
import distutils.log
from setuptools import find_packages, setup


class CleanAllCommand(distutils.cmd.Command):
    """Docstring for public class."""

    description = "remove extra build files"
    user_options = []
    dirname = os.path.dirname(os.path.realpath(__file__))

    def initialize_options(self):
        """Docstring for public method."""
        pass

    def finalize_options(self):
        """Docstring for public method."""
        pass

    def run(self):
        """Docstring for public method."""
        targets = [
            ".cache",
            ".coverage.py27",
            ".coverage.py36",
            ".tox",
            "coverage-html.py27",
            "coverage-html.py36",
            "consoleme.egg-info",
            "consoleme/__pycache__",
            "test/__pycache__",
        ]
        for t in targets:
            path = os.path.join(self.dirname, t)
            if os.path.isfile(path):
                self.announce(f"removing file: {path}", level=distutils.log.INFO)
                os.remove(path)
            elif os.path.isdir(path):
                self.announce(f"removing directory: {path}", level=distutils.log.INFO)
                rmtree(path)


setup(
    name="consoleme",
    author="Netflix Security",
    author_email="consoleme-maintainers@netflix.com",
    description="A central control plane for AWS permissions and access",
    keywords="consoleme",
    url="https://github.com/Netflix/ConsoleMe",
    python_requires=">=3.8",
    setup_requires=["setupmeta"],
    packages=find_packages(exclude=("tests",)),
    entry_points={
        "console_scripts": [
            "consoleme = consoleme.__main__:init",
        ],
        "consoleme.plugins": [
            "default_config = consoleme.default_plugins.plugins.config.config:Config",
            "default_auth = consoleme.default_plugins.plugins.auth.auth:Auth",
            "default_aws = consoleme.default_plugins.plugins.aws.aws:Aws",
            "default_celery_tasks = consoleme.default_plugins.plugins.celery_tasks.celery_tasks:internal_schedule",
            "default_celery_tasks_functions = consoleme.default_plugins.plugins.celery_tasks.celery_tasks:CeleryTasks",
            "default_metrics = consoleme.default_plugins.plugins.metrics.metrics:Metric",
            "default_policies = consoleme.default_plugins.plugins.policies.policies:Policies",
            "default_group_mapping = consoleme.default_plugins.plugins.group_mapping.group_mapping:GroupMapping",
            "default_internal_routes = consoleme.default_plugins.plugins.internal_routes.internal_routes:InternalRoutes",
        ],
    },
    cmdclass={"cleanall": CleanAllCommand},
    include_package_data=True,
    versioning="dev",
    zip_safe=False,
)



from typing import List

import sentry_sdk

from consoleme.config import config
from consoleme.handlers.base import BaseAPIV2Handler
from consoleme.lib.generic import is_in_group
from consoleme.lib.notifications.models import (
    ConsoleMeNotificationUpdateAction,
    ConsoleMeNotificationUpdateRequest,
    ConsoleMeUserNotification,
    GetNotificationsForUserResponse,
)
from consoleme.lib.plugins import get_plugin_by_name
from consoleme.lib.v2.notifications import (
    fetch_notification,
    get_notifications_for_user,
    write_notification,
)
from consoleme.models import Status2, WebResponse

stats = get_plugin_by_name(config.get("plugins.metrics", "default_metrics"))()
log = config.get_logger()


class NotificationsHandler(BaseAPIV2Handler):
    """
    A web handler for serving, updating, and (in the future) creating notifications. Current notifications are based
    around policy generation from CloudTrail errors.
    """

    async def get(self):
        try:
            notification_response: GetNotificationsForUserResponse = (
                await get_notifications_for_user(self.user, self.groups)
            )
            notifications: List[
                ConsoleMeUserNotification
            ] = notification_response.notifications
            response = WebResponse(
                status="success",
                status_code=200,
                data={
                    "unreadNotificationCount": notification_response.unread_count,
                    "notifications": notifications,
                },
            )
            self.write(response.json())
        except Exception as e:
            sentry_sdk.capture_exception()
            self.set_status(500)
            response = WebResponse(
                status=Status2.error, status_code=500, errors=[str(e)], data=[]
            )
            self.write(response.json())
            return

    async def post(self):
        # Create a notification
        raise NotImplementedError()

    async def put(self):
        """
        Allows an "authorized user" (Any user the notification is intended for) to mark the notification as read/unread
        or hidden/unhidden for themselves or all other notification recipients

        :return:
        """
        change = ConsoleMeNotificationUpdateRequest.parse_raw(self.request.body)
        errors = []

        for untrusted_notification in change.notifications:
            notification = await fetch_notification(
                untrusted_notification.predictable_id
            )
            if not notification:
                errors.append("Unable to find matching notification")
                continue
            authorized = is_in_group(
                self.user, self.groups, notification.users_or_groups
            )
            if not authorized:
                errors.append(
                    f"Unauthorized because user is not associated with notification: {notification.predictable_id}"
                )
                continue
            if (
                change.action
                == ConsoleMeNotificationUpdateAction.toggle_read_for_current_user
            ):
                if self.user in notification.read_by_users:
                    # Mark as unread
                    notification.read_by_users.remove(self.user)
                else:
                    # Mark as read
                    notification.read_by_users.append(self.user)
            elif (
                change.action
                == ConsoleMeNotificationUpdateAction.toggle_read_for_all_users
            ):
                # Mark or unmark notification as `read_by_all`.  If unmarked,
                # ConsoleMe will fall back to `notification.read_by_user` to determine if
                # a given user has read the notification
                notification.read_by_all = not notification.read_by_all
            elif (
                change.action
                == ConsoleMeNotificationUpdateAction.toggle_hidden_for_current_user
            ):
                if self.user in notification.hidden_for_users:
                    # Unmark as hidden
                    notification.hidden_for_users.remove(self.user)
                else:
                    # Mark as hidden
                    notification.hidden_for_users.append(self.user)
            elif (
                change.action
                == ConsoleMeNotificationUpdateAction.toggle_hidden_for_all_users
            ):
                # Mark or unmark as "Hidden for all users". If unmarked, falls back to `hidden_for_users.read_by_user`
                # to determine whether to show the notification to a given user
                notification.hidden_for_all = not notification.hidden_for_all
            else:
                raise Exception("Unknown or unsupported change action.")
            await write_notification(notification)
        try:
            # Retrieve and return updated notifications for user
            notification_response: GetNotificationsForUserResponse = (
                await get_notifications_for_user(
                    self.user, self.groups, force_refresh=True
                )
            )
            notifications: List[
                ConsoleMeUserNotification
            ] = notification_response.notifications
            response = WebResponse(
                status="success",
                status_code=200,
                data={
                    "unreadNotificationCount": notification_response.unread_count,
                    "notifications": notifications,
                },
            )
            self.write(response.json())
        except Exception as e:
            sentry_sdk.capture_exception()
            self.set_status(500)
            response = WebResponse(
                status=Status2.error, status_code=500, errors=[str(e)], data=[]
            )
            self.write(response.json())
            return



from logging import ERROR, Handler, getLogger

import ujson as json
from elasticsearch import Elasticsearch
from elasticsearch.exceptions import ConnectionTimeout


class ESHandler(Handler):
    def __init__(self, es, index):
        super().__init__()
        self.es = Elasticsearch(es)
        self.index = index
        getLogger("elasticsearch").setLevel(ERROR)
        getLogger("elasticsearch.trace").setLevel(ERROR)
        getLogger("urllib3").setLevel(ERROR)

    def emit(self, record):
        if isinstance(record.msg, dict):
            record.msg = json.dumps(record.msg)

        log_entry = self.format(record)

        # Push action to ES
        try:
            return self.es.index(
                index=self.index, doc_type="python_log", body=log_entry
            )
        except ConnectionTimeout:
            return False



"""
This module controls defines internal-only celery tasks and their applicable schedules. These will be combined with
the external tasks

"""
import json
import os
from datetime import timedelta

from asgiref.sync import async_to_sync
from celery import Celery

from consoleme.config import config
from consoleme.lib.json_encoder import SetEncoder
from consoleme.lib.redis import RedisHandler
from consoleme.lib.timeout import Timeout

region = config.region
red = async_to_sync(RedisHandler().redis)()

app = Celery(
    "tasks",
    broker=config.get(
        f"celery.broker.{config.region}",
        config.get("celery.broker.global", "redis://127.0.0.1:6379/1"),
    ),
    backend=config.get(
        f"celery.backend.{config.region}",
        config.get("celery.broker.global", "redis://127.0.0.1:6379/2"),
    ),
)

if config.get("redis.use_redislite"):
    import tempfile

    import redislite

    redislite_db_path = os.path.join(
        config.get("redis.redislite.db_path", tempfile.NamedTemporaryFile().name)
    )
    redislite_client = redislite.Redis(redislite_db_path)
    redislite_socket_path = f"redis+socket://{redislite_client.socket_file}"
    app = Celery(
        "tasks",
        broker=f"{redislite_socket_path}?virtual_host=1",
        backend=f"{redislite_socket_path}?virtual_host=2",
    )

app.conf.result_expires = config.get("celery.result_expires", 60)
app.conf.worker_prefetch_multiplier = config.get("celery.worker_prefetch_multiplier", 4)
app.conf.task_acks_late = config.get("celery.task_acks_late", True)

if config.get("celery.purge") and not config.get("redis.use_redislite"):
    # Useful to clear celery queue in development
    with Timeout(seconds=5, error_message="Timeout: Are you sure Redis is running?"):
        app.control.purge()


@app.task(soft_time_limit=600)
def cache_application_information():
    """
    This task retrieves application information from configuration. You may want to override this function to
    utilize your organization's CI/CD pipeline for this information.
    :return:
    """
    apps_to_roles = {}
    for k, v in config.get("application_settings", {}).items():
        apps_to_roles[k] = v.get("roles", [])

    red.set(
        config.get("celery.apps_to_roles.redis_key", "APPS_TO_ROLES"),
        json.dumps(apps_to_roles, cls=SetEncoder),
    )


@app.task
def task_1():
    """
    This task demonstrates how you can implement your own internal celery tasks to run on schedule or on demand.
    :return:
    """
    pass


schedule = timedelta(seconds=1800)

internal_schedule = {
    "task1": {
        "task": "consoleme.default_plugins.plugins.celery_tasks.celery_tasks.task_1",
        "options": {"expires": 4000},
        "schedule": schedule,
    },
    "cache_application_information": {
        "task": "consoleme.default_plugins.plugins.celery_tasks.celery_tasks.cache_application_information",
        "options": {"expires": 4000},
        "schedule": schedule,
    },
}


def init():
    """Initialize the Celery Tasks plugin."""
    return internal_schedule



import sys

from consoleme.config import config
from consoleme.handlers.base import BaseHandler
from consoleme.lib.web import handle_generic_error_response
from consoleme.models import WebResponse

log = config.get_logger()


class LogOutHandler(BaseHandler):
    async def get(self):
        log_data = {
            "function": f"{__name__}.{self.__class__.__name__}.{sys._getframe().f_code.co_name}",
            "message": "Attempting to log out user",
            "user-agent": self.request.headers.get("User-Agent"),
            "ip": self.ip,
        }
        if not config.get("auth.set_auth_cookie"):
            await handle_generic_error_response(
                self,
                "Unable to log out",
                [
                    (
                        "Configuration value `auth.set_auth_cookie` is not enabled. "
                        "ConsoleMe isn't able to delete an auth cookie if setting auth "
                        "cookies is not enabled."
                    )
                ],
                400,
                "logout_failure",
                log_data,
            )
            return
        cookie_name: str = config.get("auth_cookie_name", "consoleme_auth")
        if not cookie_name:
            await handle_generic_error_response(
                self,
                "Unable to log out",
                [
                    (
                        "Configuration value `auth_cookie_name` is not set. "
                        "ConsoleMe isn't able to delete an auth cookie if the auth cookie name "
                        "is not known."
                    )
                ],
                400,
                "logout_failure",
                log_data,
            )
            return
        self.clear_cookie(cookie_name)

        extra_auth_cookies: list = config.get("auth.extra_auth_cookies", [])
        for cookie in extra_auth_cookies:
            self.clear_cookie(cookie)

        redirect_url: str = config.get("auth.logout_redirect_url", "/")
        res = WebResponse(
            status="redirect",
            redirect_url=redirect_url,
            status_code=200,
            reason="logout_redirect",
            message="User has successfully logged out. Redirecting to landing page",
        )
        log.debug({**log_data, "message": "Successfully logged out user."})
        self.write(res.json())



import time
from hashlib import sha256
from typing import Dict, List, Optional

import ujson as json
from policy_sentry.querying.actions import get_actions_with_access_level

from consoleme.config import config
from consoleme.exceptions.exceptions import (
    InvalidRequestParameter,
    MissingConfigurationValue,
)
from consoleme.lib.account_indexers import get_account_id_to_name_mapping
from consoleme.lib.aws import minimize_iam_policy_statements
from consoleme.lib.defaults import SELF_SERVICE_IAM_DEFAULTS
from consoleme.lib.generic import generate_random_string, iterate_and_format_dict
from consoleme.lib.plugins import get_plugin_by_name
from consoleme.models import (
    ChangeGeneratorModel,
    ChangeGeneratorModelArray,
    ChangeModelArray,
    CrudChangeGeneratorModel,
    InlinePolicyChangeModel,
    PolicyModel,
    ResourceModel,
    Status,
)

group_mapping = get_plugin_by_name(
    config.get("plugins.group_mapping", "default_group_mapping")
)()
ALL_ACCOUNTS = None

self_service_iam_config: dict = config.get(
    "self_service_iam", SELF_SERVICE_IAM_DEFAULTS
)


async def _generate_policy_statement(
    actions: List, resources: List, effect: str, condition: Dict
) -> Dict:
    """
    Generates an IAM policy resource given actions, effects, resources, and conditions
    :param actions: a List of actions
    :param resources: a List of AWS resource ARNs or wildcards
    :param effect: an Effect (Allow|Deny)
    :return:
    """

    policy_statement = {
        "Action": list(set(actions)),
        "Effect": effect,
        "Resource": list(set(resources)),
    }
    if condition:
        policy_statement["Condition"] = condition
    return policy_statement


async def _generate_policy_sid(user: str) -> str:
    """
    Generate a unique SID identifying the user and time of the change request.

    :param user: User's e-mail address
    :return: policy SID string
    """
    user_stripped = user.split("@")[0]
    # Strip out any special characters from username
    user_stripped = "".join(e for e in user_stripped if e.isalnum())
    random_string = await generate_random_string()
    return f"cm{user_stripped}{int(time.time())}{random_string}"


async def generate_policy_name(
    policy_name: str, user: str, expiration_date: Optional[int] = None
) -> str:
    """
    Generate a unique policy name identifying the user and time of the change request.

    :param policy_name: A predefined policy name that will override the generated one, if it exists
    :param user: User's e-mail address
    :return: policy name string
    """
    temp_policy_prefix = config.get("policies.temp_policy_prefix", "cm_delete-on")
    if policy_name:
        return policy_name
    user_stripped = user.split("@")[0]
    random_string = await generate_random_string()
    if expiration_date:
        return (
            f"{temp_policy_prefix}_{expiration_date}_{user_stripped}_{int(time.time())}"
        )
    return f"cm_{user_stripped}_{int(time.time())}_{random_string}"


async def _generate_inline_policy_model_from_statements(
    statements: List[Dict],
) -> PolicyModel:
    """
    given a list of policy statements, generate a policy
    :param statements: List[Dict] - list of policy statements
    :return:
    """
    policy_document = {"Version": "2012-10-17", "Statement": statements}
    policy_string = json.dumps(policy_document)
    policy_input = {
        "policy_document": policy_document,
        "policy_sha256": sha256(policy_string.encode()).hexdigest(),
    }
    return PolicyModel(**policy_input)


async def _generate_inline_policy_change_model(
    principal: str,
    resources: List[ResourceModel],
    statements: List[Dict],
    user: str,
    is_new: bool = True,
    policy_name: Optional[str] = None,
) -> InlinePolicyChangeModel:
    """
    Generates an inline policy change model.

    :param principal: principal associated with the InlinePolicyChangeModel
    :param resources: Resource ARNs (or wildcards) of the resources associated with the InlinePolicyChangeModel
    :param statements: A list of AWS IAM policy statement dictionaries
    :param user: User e-mail address
    :param is_new: Boolean representing if we're creating a new policy or updating an existing policy
    :param policy_name: Optional policy name. If not provided, one will be generated
    :return: InlinePolicyChangeModel
    """
    policy_name = await generate_policy_name(policy_name, user)
    policy_document = await _generate_inline_policy_model_from_statements(statements)
    change_details = {
        "change_type": "inline_policy",
        "principal": principal,
        "resources": resources,
        "policy_name": policy_name,
        "new": is_new,
        "policy": policy_document,
        "status": Status.not_applied,
    }
    return InlinePolicyChangeModel(**change_details)


async def _get_policy_sentry_access_level_actions(
    service: str, access_levels: List[str]
) -> List[str]:
    """Use policy_sentry to get actions corresponding to AWS service and access_levels.
    TODO(psanders): Move this to a more sensible module

    :param resource_arn: Resource ARN (or wildcards) of the resource associated with the change
    :param access_levels: a list of CRUD operations to generate IAM policy statmeents from
    :return: actions: A list of IAM policy actions
    """
    actions: List[str] = []
    for level in access_levels:
        actions += get_actions_with_access_level(service, level)
    return actions


async def _get_actions_from_groups(
    action_groups: List[str], permissions_map: Dict[List, Dict[str, List[str]]]
) -> List[str]:
    """Get actions based on "groups" defined in permissions_map
    TODO(psanders): Move this to a more sensible module

    :param action_groups: A list of requested CRUD operations to convert into IAM actions
    :param permissions_map: A mapping of actions associated with the resource type, usually from configuration
    :return: actions: A list of IAM policy actions
    """
    actions: List[str] = []
    for ag in action_groups:
        for pm in permissions_map:
            if pm["name"] == ag:
                actions += pm.get("permissions", [])
    if not actions:
        raise InvalidRequestParameter(
            f"One or more of the passed actions is invalid for the generator type: {action_groups}"
        )
    return actions


async def _generate_s3_inline_policy_statement_from_mapping(
    generator: ChangeGeneratorModel,
) -> Dict:
    """
    Generates an S3 inline policy statement from a ChangeGeneratorModel. S3 is an edge case, thus it gets a
    unique function for this purpose. We need to consider the resource ARN and prefix.

    :param generator: ChangeGeneratorModel
    :return: policy_statement: A dictionary representing an inline policy statement.
    """
    permissions_map = (
        self_service_iam_config.get("permissions_map", {})
        .get("s3", {})
        .get("action_map")
    )
    if not permissions_map:
        raise MissingConfigurationValue(
            "Unable to find applicable action map configuration."
        )
    action_group_actions: List[str] = []
    resource_arns = []
    effect = generator.effect
    condition = generator.condition

    if isinstance(generator.resource_arn, str):
        generator.resource_arn = [generator.resource_arn]
    # Handle the bucket ARNs
    for arn in generator.resource_arn:
        if not arn.startswith("arn:aws:s3:::"):
            arn = f"arn:aws:s3:::{arn}"
        resource_arns.append(arn)

        # Make sure prefix starts with "/"
        if not generator.bucket_prefix.startswith("/"):
            generator.bucket_prefix = f"/{generator.bucket_prefix}"

        resource_arns.append(f"{arn}{generator.bucket_prefix}")

    for action in generator.action_groups:
        action_group_actions.append(action)
    actions = await _get_actions_from_groups(action_group_actions, permissions_map)
    if generator.extra_actions:
        actions.extend(generator.extra_actions)
    return await _generate_policy_statement(actions, resource_arns, effect, condition)


async def _generate_condition_with_substitutions(generator: ChangeGeneratorModel):
    """
    Generates a condition with substitutions if they are needed.

    :param generator:
    :return:
    """
    condition: Optional[Dict] = generator.condition
    if isinstance(condition, dict):
        condition = await iterate_and_format_dict(condition, generator.dict())
    return condition


async def _generate_inline_policy_statement_from_mapping(
    generator: ChangeGeneratorModel,
) -> Dict:
    """
    Generates an inline policy statement given a ChangeGeneratorModel from a action mapping stored in configuration.

    :param generator: ChangeGeneratorModel
    :return: policy_statement: A dictionary representing an inline policy statement.
    """
    generator_type = generator.generator_type
    if not isinstance(generator_type, str):
        generator_type = generator.generator_type.value
    permissions_map = (
        self_service_iam_config.get("permissions_map", {})
        .get(generator_type, {})
        .get("action_map")
    )
    if not permissions_map:
        raise MissingConfigurationValue(
            f"Unable to find applicable action map configuration for {generator_type}."
        )

    action_group_actions: List[str] = []
    if isinstance(generator.resource_arn, str):
        generator.resource_arn = [generator.resource_arn]
    resource_arns = generator.resource_arn
    effect = generator.effect

    for action in generator.action_groups:
        # TODO: Seems like a datamodel bug when we don't have a enum defined for an array type, but I need to access
        # this as a string sometimes
        if isinstance(action, str):
            action_group_actions.append(action)
        else:
            action_group_actions.append(action)
    actions = await _get_actions_from_groups(action_group_actions, permissions_map)
    if generator.extra_actions:
        actions.extend(generator.extra_actions)
    condition: Optional[Dict] = await _generate_condition_with_substitutions(generator)
    return await _generate_policy_statement(actions, resource_arns, effect, condition)


async def _generate_inline_policy_statement_from_policy_sentry(
    generator: CrudChangeGeneratorModel,
) -> Dict:
    """
    Generates an inline policy statement given a ChangeGeneratorModel from a action mapping provided by policy
    sentry.

    :param generator: ChangeGeneratorModel
    :return: policy_statement: A dictionary representing an inline policy statement.
    """
    permissions_map = (
        self_service_iam_config.get("permissions_map", {})
        .get("crud_lookup", {})
        .get("action_map")
    )
    if not permissions_map:
        raise MissingConfigurationValue(
            "Unable to find applicable action map configuration."
        )
    access_level_actions: List[str] = []
    for access in generator.action_groups:
        for pm in permissions_map:
            if pm["name"] == access:
                access_level_actions += pm.get("permissions")
    actions = await _get_policy_sentry_access_level_actions(
        generator.service_name, access_level_actions
    )
    if generator.extra_actions:
        actions.extend(generator.extra_actions)
    if isinstance(generator.resource_arn, str):
        generator.resource_arn = [generator.resource_arn]
    return await _generate_policy_statement(
        actions, generator.resource_arn, generator.effect, generator.condition
    )


async def _generate_inline_iam_policy_statement_from_change_generator(
    change: ChangeGeneratorModel,
) -> Dict:
    """
    Generates an inline policy statement from a ChangeGeneratorModel.
    :param change: ChangeGeneratorModel
    :return: policy_statement: A dictionary representing an inline policy statement.
    """
    generator_type = change.generator_type
    if not isinstance(generator_type, str):
        generator_type = change.generator_type.value
    if generator_type == "s3":
        policy = await _generate_s3_inline_policy_statement_from_mapping(change)
    elif generator_type == "crud_lookup":
        policy = await _generate_inline_policy_statement_from_policy_sentry(change)
    else:
        policy = await _generate_inline_policy_statement_from_mapping(change)

    # Honeybee supports restricting policies to certain accounts.
    if change.include_accounts:
        policy["IncludeAccounts"] = change.include_accounts
    if change.exclude_accounts:
        policy["ExcludeAccounts"] = change.exclude_accounts
    return policy


async def _attach_sids_to_policy_statements(
    inline_iam_policy_statements: List[Dict], user: str
) -> List[Dict]:
    """
    Generates and attaches Sids to each policy statement if the statement does not already have a Sid.

    :param inline_iam_policy_statements: A list of IAM policy statement dictionaries
    :param user: The acting user's email address
    :return: A list of IAM policy statement dictionaries with Sid entries for each
    """
    for statement in inline_iam_policy_statements:
        if not statement.get("Sid"):
            statement["Sid"] = await _generate_policy_sid(user)
    return inline_iam_policy_statements


async def _generate_resource_model_from_arn(arn: str) -> Optional[ResourceModel]:
    """
    Generates a ResourceModel from a Resource ARN

    :param arn: AWS resource identifier
    :return: ResourceModel
    """
    try:
        account_id = arn.split(":")[4]
        resource_type = arn.split(":")[2]
        region = arn.split(":")[3]
        name = arn.split(":")[5].split("/")[-1]
        if not region:
            region = "global"
        global ALL_ACCOUNTS
        if not ALL_ACCOUNTS:
            ALL_ACCOUNTS = await get_account_id_to_name_mapping()
        account_name = ALL_ACCOUNTS.get(account_id, "")

        return ResourceModel(
            arn=arn,
            name=name,
            account_id=account_id,
            account_name=account_name,
            resource_type=resource_type,
            region=region,
        )
    except IndexError:
        # Resource is not parsable or a wildcard.
        return


async def generate_change_model_array(
    changes: ChangeGeneratorModelArray,
) -> ChangeModelArray:
    """
    Compiles a ChangeModelArray which includes all of the AWS policies required to satisfy the
    ChangeGeneratorModelArray request.

    :param changes: ChangeGeneratorModelArray
    :return: ChangeModelArray
    """

    change_models = []
    inline_iam_policy_statements: List[Dict] = []
    primary_principal = None
    primary_user = None
    resources = []

    for change in changes.changes:
        # Enforce a maximum of one user per ChangeGeneratorModelArray (aka Policy Request)
        if not primary_user:
            primary_user = change.user
        if primary_user != change.user:
            raise InvalidRequestParameter(
                "All changes associated with request must be associated with the same user."
            )

        # Enforce a maximum of one principal ARN per ChangeGeneratorModelArray (aka Policy Request)
        if not primary_principal:
            primary_principal = change.principal
        if primary_principal != change.principal:
            raise InvalidRequestParameter(
                "We only support making changes to a single principal ARN per request."
            )

        if change.generator_type == "custom_iam":
            inline_policies = change.policy["Statement"]
            if isinstance(inline_policies, dict):
                inline_policies = [inline_policies]
        else:
            # Generate inline policy for the change, if applicable
            inline_policies = [
                await _generate_inline_iam_policy_statement_from_change_generator(
                    change
                )
            ]
        for inline_policy in inline_policies:
            # Inline policies must have Action|NotAction, Resource|NotResource, and an Effect
            if inline_policy and (
                (not inline_policy.get("Action") and not inline_policy.get("NotAction"))
                or (
                    not inline_policy.get("Resource")
                    and not inline_policy.get("NotResource")
                )
                or not inline_policy.get("Effect")
            ):
                raise InvalidRequestParameter(
                    f"Generated inline policy is invalid. Double-check request parameter: {inline_policy}"
                )
            if inline_policy and change.resource_arn:
                # TODO(ccastrapel): Add more details to the ResourceModel when we determine we can use it for something.
                if isinstance(change.resource_arn, str):
                    change.resource_arn = [change.resource_arn]
                for arn in change.resource_arn:
                    resource_model = await _generate_resource_model_from_arn(arn)
                    # If the resource arn is actually a wildcard, we might not have a valid resource model
                    if resource_model:
                        resources.append(resource_model)
            if inline_policy:
                inline_iam_policy_statements.append(inline_policy)

        # TODO(ccastrapel): V2: Generate resource policies for the change, if applicable

    # Minimize the policy statements to remove redundancy
    inline_iam_policy_statements = await minimize_iam_policy_statements(
        inline_iam_policy_statements
    )
    # Attach Sids to each of the statements that will help with identifying who made the request and when.
    inline_iam_policy_statements = await _attach_sids_to_policy_statements(
        inline_iam_policy_statements, primary_user
    )
    # TODO(ccastrapel): Check if the inline policy statements would be auto-approved and supply that context
    inline_iam_policy_change_model = await _generate_inline_policy_change_model(
        primary_principal, resources, inline_iam_policy_statements, primary_user
    )
    change_models.append(inline_iam_policy_change_model)
    return ChangeModelArray.parse_obj({"changes": change_models})



import hashlib
import os
import urllib.parse
from typing import List

from consoleme.lib.aws_secret_manager import get_aws_secret


def split_s3_path(s3_path):
    path_parts = s3_path.replace("s3://", "").split("/")
    b = path_parts.pop(0)
    k = "/".join(path_parts)
    return b, k


class Config:
    @staticmethod
    def get_config_location():
        config_location = os.environ.get("CONFIG_LOCATION")
        default_save_location = f"{os.curdir}/consoleme.yaml"
        if config_location:
            if config_location.startswith("s3://"):
                import boto3

                client = boto3.client("s3")
                bucket, key = split_s3_path(config_location)
                obj = client.get_object(Bucket=bucket, Key=key)
                config_data = obj["Body"].read()
                with open(default_save_location, "w") as f:
                    f.write(config_data.decode())
            elif config_location.startswith("AWS_SECRETS_MANAGER:"):
                secret_name = "".join(config_location.split("AWS_SECRETS_MANAGER:")[1:])
                config_data = get_aws_secret(
                    secret_name, os.environ.get("EC2_REGION", "us-east-1")
                )
                with open(default_save_location, "w") as f:
                    f.write(config_data)
            else:
                return config_location
        config_locations: List[str] = [
            default_save_location,
            os.path.expanduser("~/.config/consoleme/config.yaml"),
            "/etc/consoleme/config/config.yaml",
            "example_config/example_config_development.yaml",
        ]
        for loc in config_locations:
            if os.path.exists(loc):
                return loc
        raise Exception(
            "Unable to find ConsoleMe's configuration. It either doesn't exist, or "
            "ConsoleMe doesn't have permission to access it. Please set the CONFIG_LOCATION environment variable "
            "to the path of the configuration, or to an s3 location with your configuration"
            "(i.e: s3://YOUR_BUCKET/path/to/config.yaml). Otherwise, ConsoleMe will automatically search for the"
            f"configuration in these locations: {', '.join(config_locations)}"
        )

    @staticmethod
    def internal_functions(cfg=None):
        cfg = cfg or {}
        pass

    @staticmethod
    def is_contractor(user):
        return False

    @staticmethod
    def get_employee_photo_url(user):
        from consoleme.config import config

        # Try to get a custom employee photo url by formatting a string provided through configuration

        custom_employee_photo_url = config.get(
            "get_employee_photo_url.custom_employee_url", ""
        ).format(user=user)
        if custom_employee_photo_url:
            return custom_employee_photo_url

        # Fall back to Gravatar
        gravatar_url = (
            "https://www.gravatar.com/avatar/"
            + hashlib.md5(user.lower().encode("utf-8")).hexdigest()  # nosec
            + "?"
        )
        gravatar_url += urllib.parse.urlencode({"d": "mp"})
        return gravatar_url

    @staticmethod
    def get_employee_info_url(user):
        return None


def init():
    """Initialize the Config plugin."""
    return Config()
